/*
 * MPlayer GUI for Win32
 * Copyright (C) 2003 Sascha Sommer <saschasommer@freenet.de>
 * Copyright (C) 2006 Erik Augustson <erik_27can@yahoo.com>
 * Copyright (C) 2006 Gianluigi Tiesi <sherpya@netfarm.it>
 * Copyright (C) 2008-2010 Fabien Coeurjoly <fabien.coeurjoly@wanadoo.fr>
 *
 * This file is part of MPlayer.
 *
 * MPlayer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MPlayer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with MPlayer; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <unistd.h>
#include <proto/dos.h>
#include <intuition/intuition.h>
#include <clib/debug_protos.h>
#include <clib/macros.h>
#include <sys/stat.h>
#ifdef CONFIG_LIBCDIO
#include <cdio/cdio.h>
#endif

#include "access_mpcontext.h"
#include "codec-cfg.h"
#include "command.h"
#include "help_mp.h"
#include "playtree.h"
#include "m_config.h"
#include "m_option.h"
#include "mixer.h"
#include "mp_msg.h"
#include "mp_core.h"
#include "mpcommon.h"
#include "mplayer.h"
#include "path.h"
#include "sub/sub.h"
#include "input/input.h"
#include "libao2/audio_out.h"
#include "libmpcodecs/vd.h"
#include "libmpdemux/demuxer.h"
#include "libmpdemux/stheader.h"
#include "libvo/video_out.h"
#include "libvo/cgx_common.h"
#include "stream/stream.h"
#ifdef CONFIG_DVDREAD
#include "stream/stream_dvd.h"
#include <dvdnav/dvdnav.h>
#include "stream/stream_dvdnav.h"
#endif
#include "libass/ass.h"

#include "morphos_stuff.h"
#include "gui/interface.h"
#include "gui.h"
#include "thread.h"

#define D(x) ;

extern void dvd_read_name(char *name, char *serial, const char *device);

extern int quiet;

extern int abs_seek_pos;
extern float rel_seek_secs;
extern int vcd_track;
extern af_cfg_t af_cfg;
extern mp_osd_obj_t* vo_osd_list;
extern char **sub_name;
extern struct Screen * My_Screen;

char cfg_dvd_device[MAX_PATH];

static char dvdname[MAX_PATH];
char windowtitle[256];

int guiSetEventParam = 0;
int guiWinID = 0;
char *skinName = NULL;
char *codecname = NULL;
int mplGotoTheNext = 1;
gui_t *mygui = NULL;
int gui_initialized = FALSE;
int gui_failed = 0;
const ao_functions_t *audio_out = NULL;
const vo_functions_t *video_out = NULL;
mixer_t *mixer = NULL;

static struct Thread * guithread = NULL;

void capitalize(char *filename)
{
    unsigned int i;
    BOOL cap = TRUE;
    for (i=0; i < strlen(filename); i++)
    {
        if (cap)
        {
            cap = FALSE;
            filename[i] = toupper(filename[i]);
        }
        else if (filename[i] == ' ')
            cap = TRUE;
        else
            filename[i] = tolower(filename[i]);
    }
}

/* test for playlist files, no need to specify -playlist on the commandline.
 * add any conceivable playlist extensions here.
 * - Erik
 */
int parse_filename(char *file, play_tree_t *playtree, m_config_t *mconfig, int clear)
{
    if(clear)
	{
        mygui->playlist->clear_playlist(mygui->playlist);
	}

    if(strstr(file, ".m3u") || strstr(file, ".pls"))
    {
        playtree = parse_playlist_file(file);
		if(playtree)
		{
			import_playtree_playlist_into_gui(playtree, mconfig);
			return 1;
		}
	}
    return 0;
}

STRPTR get_title(void)
{
	STRPTR titleptr = NULL;

	if(guiIntfStruct.mpcontext && guiIntfStruct.Filename)
	{
		STRPTR title = property_expand_string(guiIntfStruct.mpcontext, "${metadata/Title}");
		ULONG use_filepart = strstr(guiIntfStruct.Filename, "://") == NULL;

		if(title && *title)
		{
			titleptr = title;
		}
		else if(mygui->icy_info)
		{
			titleptr = mygui->icy_info;
		}
		else if(use_filepart)
		{
			titleptr = FilePart(guiIntfStruct.Filename);
		}
		else
		{
			titleptr = guiIntfStruct.Filename;
		}
	}

	return titleptr;
}

void guiUpdateICY(char * info)
{
	char * ptr_start = strstr(info, "StreamTitle='");
	char * ptr_end = NULL;

	if(ptr_start)
	{
		ptr_start += strlen("StreamTitle='");
		ptr_end = strstr(ptr_start, "'");
	}

	free(mygui->icy_info);

	if(ptr_start && ptr_end)
	{
		mygui->icy_info = malloc(ptr_end - ptr_start + 1);
		if(mygui->icy_info)
		{
			stccpy(mygui->icy_info, ptr_start, ptr_end - ptr_start + 1);
		}
	}
	else
	{
		mygui->icy_info = strdup(info);
	}

	DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_SetValues, NULL);
}

/**
 * \brief this actually creates a new list containing only one element...
 */
void gaddlist( char ***list, const char *entry)
{
    int i;

    if (*list)
    {
        for (i=0; (*list)[i]; i++) free((*list)[i]);
        free(*list);
    }

    *list = malloc(2 * sizeof(char **));
    (*list)[0] = gstrdup(entry);
    (*list)[1] = NULL;
}

char *gstrdup(const char *str)
{
    if (!str) return NULL;
    return strdup(str);
}

/**
 * \brief this replaces a string starting with search by replace.
 * If not found, replace is appended.
 */
void greplace(char ***list, char *search, char *replace)
{
    int i = 0;
    int len = (search) ? strlen(search) : 0;

    if (*list)
    {
        for (i = 0; (*list)[i]; i++)
        {
            if (search && (!strncmp((*list)[i], search, len)))
            {
                free((*list)[i]);
                (*list)[i] = gstrdup(replace);
                return;
            }
        }
    *list = realloc(*list, (i + 2) * sizeof(char *));
    }
    else
        *list = malloc(2 * sizeof(char *));

    (*list)[i] = gstrdup(replace);
    (*list)[i + 1] = NULL;
}

/* really hardcore */
void guiClearSubtitle(void)
{
    if (vo_osd_list)
    {
        int len;
        mp_osd_obj_t *osd = vo_osd_list;
        while (osd)
        {
            if (osd->type == OSDTYPE_SUBTITLE) break;
            osd = osd->next;
        }
        if (osd && osd->flags & OSDFLAG_VISIBLE)
        {
            len = osd->stride * (osd->bbox.y2 - osd->bbox.y1);
            memset(osd->bitmap_buffer, 0, len);
            memset(osd->alpha_buffer, 0, len);
        }
    }
}

void guiLoadSubtitle(char *name)
{
    if (subdata)
    {
		int i = 0;
        mp_msg(MSGT_GPLAYER, MSGL_INFO, MSGTR_DeletingSubtitles);
        sub_free(subdata);
        subdata = NULL;
        vo_sub = NULL;

		if(sub_name)
		{
			while(sub_name[i])
			{
				free(sub_name[i]);
				i++;
			}
			free(sub_name);
			sub_name = NULL;
		}

        if (vo_osd_list)
        {
            int len;
            mp_osd_obj_t *osd = vo_osd_list;
            while (osd)
            {
                if (osd->type == OSDTYPE_SUBTITLE) break;
                osd = osd->next;
            }
            if (osd && osd->flags & OSDFLAG_VISIBLE)
            {
                len = osd->stride * (osd->bbox.y2 - osd->bbox.y1);
                memset(osd->bitmap_buffer, 0, len);
                memset(osd->alpha_buffer, 0, len);
            }
        }
    }

    if (name)
    {
        mp_msg(MSGT_GPLAYER, MSGL_INFO, MSGTR_LoadingSubtitles, name);
        subdata = sub_read_file(strdup(name), guiIntfStruct.FPS);
        if (!subdata) mp_msg(MSGT_GPLAYER, MSGL_ERR, MSGTR_CantLoadSub,name);
        sub_name = (malloc(2 * sizeof(char*))); /* when mplayer will be restarted */
		if(sub_name)
		{
	        sub_name[0] = strdup(name);               /* sub_name[0] will be read */
	        sub_name[1] = NULL;
		}
    }
    update_set_of_subtitles();
}

void startplay(gui_t * gui)
{
	gui->playercontrol(evDropFile);
}

void updatedisplay(gui_t * gui)
{
	DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 3, MM_MPlayerGroup_Update, MV_MPlayerGroup_Update_All, NULL);
}

void update_windowdimensions(void)
{
	ULONG menuentry;
	switch(gui_window_dimensions)
	{
		default:
		case 0:
			menuentry = MEN_DIMENSIONS_FREE;
			break;
		case 1:
			menuentry = MEN_DIMENSIONS_ORIGINAL;
			break;
		case 2:
			menuentry = MEN_DIMENSIONS_HALF;
			break;
		case 3:
			menuentry = MEN_DIMENSIONS_DOUBLE;
			break;
	}

	DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_Dimensions, menuentry);
}

void update_playlistwindow(ULONG rebuild)
{
	DoMethod(app, MUIM_Application_PushMethod, mygui->playlistgroup, 2, MM_PlaylistGroup_Refresh, rebuild);
}

/* this function gets called by the gui to update mplayer */
void guiSetEvent(int event)
{
    if(guiIntfStruct.mpcontext)
        mixer = mpctx_get_mixer(guiIntfStruct.mpcontext);

    switch(event)
    {
        case evPlay:
        case evPlaySwitchToPause:
        case evPauseSwitchToPlay:
			mplPlay();
            break;
        case evPause:
            mplPause();
            break;
#ifdef CONFIG_DVDREAD
		case evSetDVDAngle:
		{
			 char command[64];
			 snprintf(command, sizeof(command), "switch_angle %d", guiSetEventParam);
			 guiIntfStruct.DVD.current_angle = guiSetEventParam;
			 mp_input_queue_cmd(mp_input_parse_cmd(command));
	         break;
		}
	    case evSetDVDChapter:
		{
			 char command[64];
			 snprintf(command, sizeof(command), "seek_chapter %d 1", guiSetEventParam);
			 guiIntfStruct.DVD.current_chapter = guiSetEventParam;
			 mp_input_queue_cmd(mp_input_parse_cmd(command));	          
	         break;
		}
		case evSetDVDTitle:
		{
             guiIntfStruct.DVD.current_title = guiSetEventParam;
			 goto playdvd;

	         break;
		}
        case evPlayDVD:
        {
			char buffer[MAX_PATH];

			D(kprintf("evPlayDVD\n"));
			D(kprintf("dvd_title %d dvd_chapter %d dvd_angle %d\n", dvd_title, dvd_chapter, dvd_angle));

			guiIntfStruct.DVD.current_title = 1/*dvd_title + 1*/;
playdvd:
			guiIntfStruct.DVD.current_chapter = 1/*dvd_chapter*/;
			guiIntfStruct.DVD.current_angle = 1/*dvd_angle*/;
            guiIntfStruct.DiskChanged = 1;

			if(!dvd_device) break; // warn user with a requester

			D(kprintf("-> title %d chapter %d angle %d\n", guiIntfStruct.DVD.current_title, guiIntfStruct.DVD.current_chapter, guiIntfStruct.DVD.current_angle));
			mplSetFileName(NULL, dvd_device, gui_use_dvdnav ? STREAMTYPE_DVDNAV: STREAMTYPE_DVD);

			/* DVD title */
            dvdname[0] = 0;
			buffer[0] = 0;
			dvd_read_name(buffer, NULL, dvd_device);
			capitalize(buffer);
			snprintf(dvdname, sizeof(dvdname), "DVD Movie %s", buffer[0] ? buffer : dvd_device );
            mp_msg(MSGT_GPLAYER, MSGL_V, "Opening DVD %s -> %s\n", dvd_device, dvdname);

			guiGetEvent(guiSetParameters, (char *) (gui_use_dvdnav ? STREAMTYPE_DVDNAV : STREAMTYPE_DVD));

			mygui->playlist->clear_playlist(mygui->playlist);
			mygui->playlist->add_track(mygui->playlist, filename, NULL, dvdname, 0, gui_use_dvdnav ? STREAMTYPE_DVDNAV: STREAMTYPE_DVD);

			// Rebuild playlist
			update_playlistwindow(TRUE);

			mygui->startplay(mygui);

            break;
        }
#endif
#ifdef CONFIG_LIBCDIO
        case evPlayCD:
        {
            int i;
            char track[10];
            char trackname[10];
            CdIo_t *p_cdio = cdio_open(NULL, DRIVER_UNKNOWN);
            track_t i_tracks;

            if(p_cdio == NULL) printf("Couldn't find a driver.\n");
            i_tracks = cdio_get_num_tracks(p_cdio);

			mygui->playlist->clear_playlist(mygui->playlist);
            for(i=0;i<i_tracks;i++)
            {
                sprintf(track, "cdda://%d", i+1);
                sprintf(trackname, "Track %d", i+1);
				mygui->playlist->add_track(mygui->playlist, track, NULL, trackname, 0, STREAMTYPE_VCD);
            }
            cdio_destroy(p_cdio);
			mygui->startplay(mygui);
            break;
        }
#endif
        case evDropFile:
        case evLoadPlay:
        {
			D(kprintf("evLoadPlay\n"));

            switch(guiIntfStruct.StreamType)
            {
#ifdef CONFIG_DVDREAD
				case STREAMTYPE_DVDNAV:
				case STREAMTYPE_DVD:
                {
					D(kprintf("title %d chapter %d angle %d\n", guiIntfStruct.DVD.current_title, guiIntfStruct.DVD.current_chapter, guiIntfStruct.DVD.current_angle));

                    guiIntfStruct.Title = guiIntfStruct.DVD.current_title;
                    guiIntfStruct.Chapter = guiIntfStruct.DVD.current_chapter;
                    guiIntfStruct.Angle = guiIntfStruct.DVD.current_angle;
                    guiIntfStruct.DiskChanged = 1;
                    guiGetEvent(guiCEvent, (void *) guiSetPlay);
                    break;
                }
#endif
                default:
                {
                    guiIntfStruct.FilenameChanged = guiIntfStruct.NewPlay = 1;
                    mplGotoTheNext = guiIntfStruct.Playing? 0 : 1;
                    guiGetEvent(guiCEvent, (void *) guiSetStop);
                    guiGetEvent(guiCEvent, (void *) guiSetPlay);
                    break;
               }
           }

		   // Update playlist cursor
		   update_playlistwindow(FALSE);
           break;
        }
        case evNext:
            mplNext();
            break;
        case evPrev:
            mplPrev();
            break;
        case evFullScreen:
            mp_input_queue_cmd(mp_input_parse_cmd("vo_fullscreen"));
            break;
        case evExit:
        {
            /* We are asking mplayer to exit, later it will ask us after uninit is made
               this should be the only safe way to quit */
            mp_input_queue_cmd(mp_input_parse_cmd("quit"));
            break;
        }
        case evStop:
            if(guiIntfStruct.Playing)
                guiGetEvent(guiCEvent, (void *) guiSetStop);
            break;
        case evSetMoviePosition:
        {
            rel_seek_secs = guiIntfStruct.Position / 100.0f;
            abs_seek_pos = 3;
            break;
        }
        case evForward10sec:
        {
            rel_seek_secs = 10.0f;
            abs_seek_pos = 0;
            break;
        }
        case evBackward10sec:
        {
            rel_seek_secs = -10.0f;
            abs_seek_pos = 0;
            break;
        }
		case evForward1min:
        {
			rel_seek_secs = 60.0f;
            abs_seek_pos = 0;
            break;
        }
		case evBackward1min:
        {
			rel_seek_secs = -60.0f;
            abs_seek_pos = 0;
            break;
        }
		case evForward10min:
        {
			rel_seek_secs = 600.0f;
            abs_seek_pos = 0;
            break;
        }
		case evBackward10min:
        {
			rel_seek_secs = -600.0f;
            abs_seek_pos = 0;
            break;
        }
        case evSetBalance:
        case evSetVolume:
        {
            float l,r;

            if (guiIntfStruct.Playing == 0)
                break;

            if (guiIntfStruct.Balance == 50.0f)
                mixer_setvolume(mixer, guiIntfStruct.Volume, guiIntfStruct.Volume);

            l = guiIntfStruct.Volume * ((100.0f - guiIntfStruct.Balance) / 50.0f);
            r = guiIntfStruct.Volume * ((guiIntfStruct.Balance) / 50.0f);

            if (l > guiIntfStruct.Volume) l=guiIntfStruct.Volume;
            if (r > guiIntfStruct.Volume) r=guiIntfStruct.Volume;
            mixer_setvolume(mixer, l, r);
            /* Check for balance support on mixer - there is a better way ?? */
            if (r != l)
            {
                mixer_getvolume(mixer, &l, &r);
                if (r == l)
                {
                    mp_msg(MSGT_GPLAYER, MSGL_V, "[GUI] Mixer doesn't support balanced audio\n");
                    mixer_setvolume(mixer, guiIntfStruct.Volume, guiIntfStruct.Volume);
                    guiIntfStruct.Balance = 50.0f;
                }
            }
            break;
        }
        case evMute:
        {
            mp_cmd_t * cmd = calloc(1, sizeof(*cmd));
            cmd->id=MP_CMD_MUTE;
            cmd->name=strdup("mute");
            mp_input_queue_cmd(cmd);
            break;
        }
    }
}

void mplPlay( void )
{
   D(kprintf("mplPlay\n"));

   if((!guiIntfStruct.Filename ) || (guiIntfStruct.Filename[0] == 0))
     return;

   if(guiIntfStruct.Playing > 0)
   {
       mplPause();
       return;
   }
   guiIntfStruct.NewPlay = 1;
   guiGetEvent(guiCEvent, (void *) guiSetPlay);
}

void mplPause( void )
{
   if(!guiIntfStruct.Playing) return;

   if(guiIntfStruct.Playing == 1)
   {
       mp_cmd_t * cmd = calloc(1, sizeof(*cmd));
       cmd->id=MP_CMD_PAUSE;
       cmd->name=strdup("pause");
       mp_input_queue_cmd(cmd);
   } else guiIntfStruct.Playing = 1;
}

void mplNext(void)
{
	D(kprintf("mplNext\n"));
    if(guiIntfStruct.Playing == 2) return;
    switch(guiIntfStruct.StreamType)
    {

#ifdef CONFIG_DVDREAD
        case STREAMTYPE_DVD:
		case STREAMTYPE_DVDNAV:
			if(guiIntfStruct.DVD.current_chapter == (guiIntfStruct.DVD.chapters - 1))
				return;
			guiSetEventParam = ++guiIntfStruct.DVD.current_chapter;
			mygui->playercontrol(evSetDVDChapter);

            break;
#endif

        default:

            if(mygui->playlist->current == (mygui->playlist->trackcount - 1))
                return;
			if(!mygui->playlist->tracks)
				return;
			mygui->playlist->current++;
			mplSetFileName(NULL, mygui->playlist->tracks[mygui->playlist->current]->filename,
						   mygui->playlist->tracks[mygui->playlist->current]->type);

			mygui->startplay(mygui);
            break;
    }
}

void mplPrev(void)
{
	D(kprintf("mplPrev\n"));
    if(guiIntfStruct.Playing == 2) return;
    switch(guiIntfStruct.StreamType)
    {

#ifdef CONFIG_DVDREAD
        case STREAMTYPE_DVD:
		case STREAMTYPE_DVDNAV:
			if(guiIntfStruct.DVD.current_chapter == 1)
                return;
			guiSetEventParam = --guiIntfStruct.DVD.current_chapter;
			mygui->playercontrol(evSetDVDChapter);

            break;
#endif

        default:
            if(mygui->playlist->current == 0)
                return;
			if(!mygui->playlist->tracks)
				return;
			mygui->playlist->current--;
			mplSetFileName(NULL, mygui->playlist->tracks[mygui->playlist->current]->filename,
                           mygui->playlist->tracks[mygui->playlist->current]->type);
			
			mygui->startplay(mygui);
			break;
    }
}

void mplEnd( void )
{
	D(kprintf("mplEnd\n"));

	/* Repeat current file */
	if(guiIntfStruct.Playing && gui_repeat_mode == 1)
	{
		mplGotoTheNext = 1;

		if(!mygui->playlist->tracks)
			return;

		update_playlistwindow(FALSE);
		guiGetEvent(guiCEvent, (void *) guiSetPlay);
		return;
	}

	if(!mplGotoTheNext && guiIntfStruct.Playing)
    {
        mplGotoTheNext = 1;
        return;
    }

    if(mplGotoTheNext && guiIntfStruct.Playing &&
	  (mygui->playlist->current < (mygui->playlist->trackcount - 1)))
    {
        /* we've finished this file, reset the aspect */
        if(movie_aspect >= 0)
            movie_aspect = -1;

        mplGotoTheNext = guiIntfStruct.FilenameChanged = guiIntfStruct.NewPlay = 1;

		if(mygui->playlist->tracks)
		{
			mygui->playlist->current++;
			mplSetFileName(NULL, mygui->playlist->tracks[mygui->playlist->current]->filename, mygui->playlist->tracks[mygui->playlist->current]->type);
		}

		// Update playlist cursor
		update_playlistwindow(FALSE);
    }

    if(guiIntfStruct.FilenameChanged && guiIntfStruct.NewPlay)
        return;

    guiIntfStruct.TimeSec = 0;
    guiIntfStruct.Position = 0;
    guiIntfStruct.AudioType = 0;

#ifdef CONFIG_DVDREAD
    guiIntfStruct.DVD.current_title = 1;
    guiIntfStruct.DVD.current_chapter = 1;
    guiIntfStruct.DVD.current_angle = 1;
#endif

    if (mygui->playlist->current == (mygui->playlist->trackcount - 1))
        mygui->playlist->current = 0;

	// Update playlist cursor
	update_playlistwindow(FALSE);

	if(!mygui->embedded)
		fullscreen = 0; // See what it really implies

	/* Loop over playlist, or just stop */
	if(gui_repeat_mode == 2)
	{
		guiGetEvent(guiCEvent, (void *) guiSetPlay);
	}
	else if(gui_repeat_mode == 3)
	{
		exit_player(EXIT_NONE);
	}
	else
	{
	    guiGetEvent(guiCEvent, (void *) guiSetStop);
		guiGetEvent(guiShowPanel, (void *) TRUE);
	}
}

void mplSetFileName(char *dir, char *name, int type)
{
	D(kprintf("mplSetFileName\n"));

    if(!name) return;
    if(!dir)
        guiSetFilename(guiIntfStruct.Filename, name)
    else
        guiSetDF(guiIntfStruct.Filename, dir, name);

	D(kprintf("guiIntfStruct.Filename = <%s> type %d\n", guiIntfStruct.Filename, type));

    guiIntfStruct.StreamType = type;
	free(guiIntfStruct.AudioFile);
	guiIntfStruct.AudioFile = NULL;
	free(guiIntfStruct.Subtitlename);
	guiIntfStruct.Subtitlename = NULL;

	/* reset subtitles */
	guiClearSubtitle();
}

void mplFullScreen( void )
{
	if(!guiIntfStruct.sh_video) return;

	if(video_out && video_out->control(VOCTRL_FULLSCREEN, 0) == VO_FALSE)
	{
		mp_msg(MSGT_GPLAYER, MSGL_INFO, "[GUI] Couldn't switch to window/fullscreen mode");
		exit_player(EXIT_ERROR);
	}
}

int stream_interrupt_cb(int timeout)
{
	return mygui->running == FALSE; 
}

void initguimembers(void)
{
	const vo_functions_t * vo;

	stccpy(cfg_dvd_device, dvd_device ? dvd_device : DEFAULT_DVD_DEVICE, sizeof(cfg_dvd_device)); /* workaround for dvddirectory to avoid overwriting original dvd_device */

	memset(mygui, 0, sizeof(*mygui));

	// Shall embedded mode be used ?
	mygui->embedded = FALSE;

	if((vo = init_best_video_out(video_driver_list)))
	{
		if(vo->info && (!strcmp(vo->info->short_name, "cgx_overlay_gui") || !strcmp(vo->info->short_name, "cgx_wpa_gui")))
		{
			mygui->embedded = TRUE;
		}
		Cgx_ControlBlanker(NULL, FALSE); /* hack to equilibrate blanker count */
        if(vo->uninit) vo->uninit();
	}

	mygui->colorkey = 0;
	mygui->startplay = startplay;
	mygui->playercontrol = guiSetEvent;
	mygui->updatedisplay = updatedisplay;
	mygui->playlist = create_playlist();
}

static int guiThread(void)
{
	int rc = 1;
	struct startupmsg * msg;

	stream_set_interrupt_callback(stream_interrupt_cb);

	/* Set cfg items */

	quiet = 1; // avoid being spammed by unwanted status messages

    if(autosync && autosync != gtkAutoSync)
    {
       gtkAutoSyncOn = 1;
       gtkAutoSync = autosync;
    }

	if(subdata) guiSetFilename(guiIntfStruct.Subtitlename, subdata->filename);

#ifdef CONFIG_ASS
    gtkASS.enabled = ass_enabled;
    gtkASS.use_margins = ass_use_margins;
    gtkASS.top_margin = ass_top_margin;
    gtkASS.bottom_margin = ass_bottom_margin;
#endif

	/* always add screenshot filter */
	m_config_set_option(mconfig, "vf-add", "screenshot");


#ifndef __amigaos4__
	if (NewGetTaskAttrs(NULL, &msg, sizeof(struct startupmsg *),
					  TASKINFOTYPE_STARTUPMSG,
					  TAG_DONE) && msg)
#else
    msg = ((struct Process *)FindTask(NULL))->pr_Task.tc_UserData;
#endif

	{
		mygui = NULL;
		gui_initialized = FALSE;
		rc = creategui();
	}

	mp_msg(MSGT_GPLAYER, MSGL_V, "[GUI] GUI thread terminated\n");

//	  guithread->finished = TRUE;

	gui_failed = rc;

#ifdef __amigaos4__
    ReplyMsg((struct Message *)msg);
#endif

    return 0;
}

void guiInit(void)
{
    memset(&guiIntfStruct, 0, sizeof(guiIntfStruct));

	if(guithread) return; /* already exists, ignore */

	/* Create gui thread */
	if(!gui_initialized)
	{
		guithread = RunThread(guiThread, THREAD_NAME, NULL, (APTR) NULL, 0);
		mp_msg(MSGT_GPLAYER, MSGL_V, "[GUI] Creating GUI Thread\n");
	}

	/* Wait until gui is created */
	while(!gui_initialized)
	{
		if(gui_failed)
		{
			WaitForThread(guithread);
			guithread = NULL;
			mp_msg(MSGT_IDENTIFY, MSGL_INFO, "[GUI] Failed to initialize GUI module\n");
			exit_player(EXIT_ERROR);
		}

		usleep(100*1000);
	}

    mp_msg(MSGT_GPLAYER, MSGL_V, "[GUI] GUI thread started.\n");
}

static void guiFree(void)
{
	free( guiIntfStruct.Filename );
	guiIntfStruct.Filename = NULL;
	free( guiIntfStruct.Subtitlename );
	guiIntfStruct.Subtitlename = NULL;
	free( guiIntfStruct.AudioFile );
	guiIntfStruct.AudioFile = NULL;
}

void guiDone(void)
{
	D(kprintf("[MAIN] guiDone()\n"));
	if(gui_initialized)
    {
		mp_msg(MSGT_GPLAYER, MSGL_V, "[GUI] Closed by main mplayer window\n");

		D(kprintf("[MAIN] signaling E to gui thread (to quit main loop)\n"));
		SignalThread(guithread);

		D(kprintf("[MAIN] signaling F to gui thread\n"));
		Signal((struct Task *) guithread->task, SIGBREAKF_CTRL_F);

		D(kprintf("[MAIN] waiting for C|F signal\n"));
		Wait(SIGBREAKF_CTRL_F|SIGBREAKF_CTRL_C);

		D(kprintf("[MAIN] waiting for thread startupmsg reply\n"));
		WaitForThread(guithread);
		guithread = NULL;
		gui_initialized = FALSE;

		D(kprintf("[MAIN] bye bye\n"));
	}

	Cgx_BlankerState(); /* just check blanker count is ok */

	if(mygui)
	{
		if(mygui->playlist)
			mygui->playlist->free_playlist(mygui->playlist);
		mygui->playlist = NULL;

		free(mygui->icy_info);

		free(mygui);
		mygui = NULL;
    }

	guiFree();

	free(dvd_device);
	dvd_device = NULL;
}

#ifdef CONFIG_DVDREAD
typedef struct
{
	dvdnav_t *dvdnav;
} dvdnavpriv_t;
#endif

/* this function gets called by mplayer to update the gui */
int guiGetEvent(int type, void *arg)
{
    stream_t *stream = (stream_t *) arg;
#ifdef CONFIG_DVDREAD
    dvd_priv_t *dvdp = (dvd_priv_t *) arg;
	dvdnavpriv_t * dvdnavp = (dvdnavpriv_t *) arg;
#endif
    if(!mygui) return 0;

    if(guiIntfStruct.mpcontext)
    {
        audio_out = mpctx_get_audio_out(guiIntfStruct.mpcontext);
        video_out = mpctx_get_video_out(guiIntfStruct.mpcontext);
        mixer = mpctx_get_mixer(guiIntfStruct.mpcontext);
        playtree = mpctx_get_playtree_iter(guiIntfStruct.mpcontext);
    }

    switch (type)
    {
        case guiSetFileFormat:
            guiIntfStruct.FileFormat = (int) arg;
            break;
        case guiSetParameters:
        {
			D(kprintf("guiSetParameters %d\n", guiIntfStruct.StreamType));
            guiGetEvent(guiSetDefaults, NULL);
            guiIntfStruct.DiskChanged = 0;
            guiIntfStruct.FilenameChanged = 0;
            guiIntfStruct.NewPlay = 0;
            switch(guiIntfStruct.StreamType)
            {
                case STREAMTYPE_PLAYLIST:
                    break;
#ifdef CONFIG_DVDREAD
				case STREAMTYPE_DVDNAV:
					stream_cache_size = -1;
				case STREAMTYPE_DVD:
                {
					char tmp[64];

					dvd_title = guiIntfStruct.DVD.current_title - 1; /* 0 */
					dvd_chapter = 1/*guiIntfStruct.DVD.current_chapter*/;
					dvd_angle = 1/*guiIntfStruct.DVD.current_angle*/;

					D(kprintf("-> guiIntfStruct.Title = %d dvd_title %d dvd_chapter %d dvd_angle %d\n", guiIntfStruct.Title, dvd_title, dvd_chapter, dvd_angle));

					if(guiIntfStruct.Title == 0)
					{
						guiIntfStruct.Title = 1;
					}

					if(gui_use_dvdnav)
					{
						sprintf(tmp,"dvdnav://%d", guiIntfStruct.Title);
					}
					else
					{
						sprintf(tmp,"dvd://%d", guiIntfStruct.Title);
					}
                    guiSetFilename(guiIntfStruct.Filename, tmp);
                    break;
                }
#endif
            }
            if(guiIntfStruct.Filename)
				filename = guiIntfStruct.Filename;
			else if(filename && guiIntfStruct.Filename)
                strcpy(guiIntfStruct.Filename, filename);
            break;
        }
        case guiSetAudioOnly:
        {
            guiIntfStruct.AudioOnly = (int) arg;
            break;
        }
        case guiSetContext:
		{
			D(kprintf("guiSetContext %p\n", arg));
            guiIntfStruct.mpcontext = (void *) arg;
            break;
		}
        case guiSetDemuxer:
		{
			D(kprintf("guiSetDemuxer %p\n", arg));
            guiIntfStruct.demuxer = (void *) arg;

			/* At this point (demuxer set), we know the stream attributes */
			DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_SetValues, arg != NULL);
			if(arg)	
			{
				// Fill stream properties
				DoMethod(app, MUIM_Application_PushMethod, mygui->propertiesgroup, 1, MM_PropertiesGroup_Update);

				// Show/Hide DVD Browser
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_ShowDVDBrowser, guiIntfStruct.StreamType == STREAMTYPE_DVDNAV || guiIntfStruct.StreamType == STREAMTYPE_DVD);

				// Chapters can also exist for other streams than DVD
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_ShowChapterBrowser, demuxer_chapter_count(guiIntfStruct.demuxer) > 1);

				// Refresh chapter/title/angle/audio/subtitle poplists/menus
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_RefreshSubtitles);
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_RefreshAudio);
			}
			break;
		}
        case guiSetValues:
        {
			D(kprintf("guiSetValues %p\n", arg));
            guiIntfStruct.sh_video = arg;
            if (arg)
            {
                sh_video_t *sh = (sh_video_t *)arg;
                codecname = sh->codec->name;
                guiIntfStruct.FPS = sh->fps;

				guiIntfStruct.MovieWidth = sh->disp_w;
				guiIntfStruct.MovieHeight = sh->disp_h;

				sub_aspect = (float)guiIntfStruct.MovieWidth / guiIntfStruct.MovieHeight;

				DoMethod(app, MUIM_Application_PushMethod, mygui->cropgroup, 1, MM_CropGroup_UpdateDimensions);
				DoMethod(app, MUIM_Application_PushMethod, mygui->scalegroup, 1, MM_ScaleGroup_UpdateDimensions);
            }

			// make sure controls are shown if this is an audio only file.
			if(!arg)
			{
				gui_show_gui = TRUE;
				gui_show_control = gui_show_status = gui_show_toolbar = TRUE;
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_ShowPanels, gui_show_gui);

				// switch back to window mode for audio only
				if(mygui->embedded)
				{
					guiGetEvent(guiSetWindowPtr, NULL);
				}
			}

			// show/hide video area if there's a video stream or not
			DoMethod(app, MUIM_Application_PushMethod, mygui->videoarea, 3, MUIM_Set, MUIA_ShowMe, mygui->embedded ? arg != NULL : FALSE);

            break;
        }
        case guiSetShVideo:
        {
			D(kprintf("guiSetShVideo %p\n", arg));

            break;
        }
        case guiSetStream:
        {
			D(kprintf("guiSetStream %d\n", stream->type));
            guiIntfStruct.StreamType = stream->type;
            switch(stream->type)
            {
#ifdef CONFIG_DVDREAD
                case STREAMTYPE_DVD:
				case STREAMTYPE_DVDNAV:
                    guiGetEvent(guiSetDVD, (char *) stream->priv);
                    break;
#endif
            }

			/* Cache */
			switch(guiIntfStruct.StreamType)
			{
				case STREAMTYPE_DVD:
					stream_cache_size = gtkCacheSizeDVD;
					break;
				case STREAMTYPE_DVDNAV:
					stream_cache_size = - 1;
					break;
				case STREAMTYPE_STREAM:
					stream_cache_size = gtkCacheSizeNet;
					break;
				default:
				case STREAMTYPE_FILE:
					stream_cache_size = gtkCacheSizeFile;
					break;
			}

			/* Title  */
			switch(guiIntfStruct.StreamType)
			{
				case STREAMTYPE_DVD:
				case STREAMTYPE_DVDNAV:
					snprintf(windowtitle, sizeof(windowtitle), "MPlayer: %s", dvdname);
					break;
				default:
				case STREAMTYPE_STREAM:
				case STREAMTYPE_FILE:
					snprintf(windowtitle, sizeof(windowtitle), "MPlayer: %s", guiIntfStruct.Filename);
					break;
			}

            if(!mygui->embedded || !fullscreen)
				DoMethod(app, MUIM_Application_PushMethod, mygui->mainwindow, 3, MUIM_Set, MUIA_Window_Title, windowtitle);

			D(kprintf("guiSetStream OK\n"));

            break;
        }
#ifdef CONFIG_DVDREAD
        case guiSetDVD:
        {
			D(kprintf("guiSetDVD\n"));

#warning "See if we couldn't use more generic demuxer functions here, really. It exposes really too much of dvd internals"

			if(guiIntfStruct.StreamType == STREAMTYPE_DVD)
			{
				D(kprintf("dvd_title %d dvd_chapter %d dvd_angle %d\n", dvd_title, dvd_chapter, dvd_angle));
	            guiIntfStruct.DVD.titles = dvdp->vmg_file->tt_srpt->nr_of_srpts;
	            guiIntfStruct.DVD.chapters = dvdp->vmg_file->tt_srpt->title[dvd_title].nr_of_ptts;
	            guiIntfStruct.DVD.angles = dvdp->vmg_file->tt_srpt->title[dvd_title].nr_of_angles;
	            guiIntfStruct.DVD.nr_of_audio_channels = dvdp->nr_of_channels;
	            memcpy(guiIntfStruct.DVD.audio_streams, dvdp->audio_streams, sizeof(dvdp->audio_streams));
	            guiIntfStruct.DVD.nr_of_subtitles = dvdp->nr_of_subtitles;
	            memcpy(guiIntfStruct.DVD.subtitles, dvdp->subtitles, sizeof(dvdp->subtitles));
	            guiIntfStruct.DVD.current_title = dvd_title + 1;
				guiIntfStruct.DVD.current_chapter = dvd_chapter + 1;
				guiIntfStruct.DVD.current_angle = dvd_angle + 1;
	            guiIntfStruct.Track = dvd_title + 1;
			}
			else if(guiIntfStruct.StreamType == STREAMTYPE_DVDNAV)
			{				 
			    uint8_t lg;
				uint16_t i, j, lang, format, id, channels;
			    int base[7] = {128, 0, 0, 0, 160, 136, 0};
			    char tmp[3];

				dvdnav_current_title_info(dvdnavp->dvdnav,
				                          &guiIntfStruct.DVD.current_title,
										  &guiIntfStruct.DVD.current_chapter);
				dvdnav_get_number_of_titles(dvdnavp->dvdnav, &guiIntfStruct.DVD.titles);
				dvdnav_get_number_of_parts(dvdnavp->dvdnav, guiIntfStruct.DVD.current_title, &guiIntfStruct.DVD.chapters);

				dvdnav_get_angle_info(dvdnavp->dvdnav,
				                      &guiIntfStruct.DVD.current_angle,
									  &guiIntfStruct.DVD.angles);

				j = 0;
				for(i=0; i<8; i++)
			    {
			      lg = dvdnav_get_audio_logical_stream(dvdnavp->dvdnav, i);
			      if(lg == 0xff) continue;

			      channels = dvdnav_audio_stream_channels(dvdnavp->dvdnav, lg);
			      if(channels == 0xFFFF)
			        channels = 2; //unknown
			      else
			        channels--;
			      lang = dvdnav_audio_stream_to_lang(dvdnavp->dvdnav, lg);
			      if(lang == 0xFFFF)
			        tmp[0] = tmp[1] = '?';
			      else
			      {
			        tmp[0] = lang >> 8;
			        tmp[1] = lang & 0xFF;
			      }
			      tmp[2] = 0;
			      format = dvdnav_audio_stream_format(dvdnavp->dvdnav, lg);
			      if(format == 0xFFFF || format > 6)
			        format = 1; //unknown
			      id = i + base[format];

				  guiIntfStruct.DVD.audio_streams[j].id = id;
				  guiIntfStruct.DVD.audio_streams[j].language = lang;
				  guiIntfStruct.DVD.audio_streams[j].channels = channels;
				  guiIntfStruct.DVD.audio_streams[j].type = format;

				  j++;
			    }

				guiIntfStruct.DVD.nr_of_audio_channels = j;

	            guiIntfStruct.Track = guiIntfStruct.DVD.current_title + 1;
			}

			D(kprintf("-> titles %d chapters %d angles %d\n", guiIntfStruct.DVD.titles, guiIntfStruct.DVD.chapters, guiIntfStruct.DVD.angles));
			D(kprintf("-> current_title %d current_chapter %d current_angle %d\n", guiIntfStruct.DVD.current_title, guiIntfStruct.DVD.current_chapter, guiIntfStruct.DVD.current_angle));

 			break;
        }
#endif
        case guiReDraw:
            break;
        case guiSetAfilter:
            guiIntfStruct.afilter = (void *) arg;
            break;
        case guiCEvent:
        {
            guiIntfStruct.Playing = (int) arg;
            switch (guiIntfStruct.Playing)
            {
                case guiSetPlay:
                {
					/* workaround: make sure dvdnav really doesn't have cache */
					if(guiIntfStruct.Filename && strstr(guiIntfStruct.Filename, "dvdnav://"))
					{
						guiIntfStruct.StreamType = STREAMTYPE_DVDNAV;
						stream_cache_size = -1;
					}

                    guiIntfStruct.Playing = 1;
                    break;
                }
                case guiSetStop:
                {
                    guiIntfStruct.Playing = 0;
                    if(movie_aspect >= 0)
                        movie_aspect = -1;
                    break;
                }
                case guiSetPause:
                    guiIntfStruct.Playing = 2;
                    break;
            }

			// Update playlist cursor
			update_playlistwindow(FALSE);

            break;
        }
        case guiIEvent:
        {
			mp_msg(MSGT_GPLAYER, MSGL_V, "cmd: %d\n", (int) arg);
            /* MPlayer asks us to quit */
            switch((int) arg)
            {
				case MP_CMD_VO_FULLSCREEN:
                    mplFullScreen();
                    break;

                case MP_CMD_QUIT:
                {
					exit_player(EXIT_NONE);
                    return 0;
                }

				case MP_CMD_GUI_HIDE:
					if(mygui->embedded && guiIntfStruct.Playing)
					{
						if(video_out)
						{
							if(video_out->control(VOCTRL_GUI_NOWINDOW, 0) == VO_FALSE)
							{
								exit_player(EXIT_NONE);
							}
						}
					}				 
					break;

				case MP_CMD_GUI_UPDATESUBTITLE:
					DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_RefreshSubtitles);
					break;

				case MP_CMD_GUI_UPDATEAUDIO:
					DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_RefreshAudio);
					break;

                case MP_CMD_GUI_STOP:
                    guiGetEvent(guiCEvent, (void *) guiSetStop);
                    break;

                case MP_CMD_GUI_PLAY:
                    guiGetEvent(guiCEvent, (void *) guiSetPlay);
                    break;

                case MP_CMD_GUI_SKINBROWSER:
                    break;

                case MP_CMD_GUI_PLAYLIST:
					DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_ShowPlaylist);
                    break;

                case MP_CMD_GUI_PREFERENCES:
                    break;

                case MP_CMD_GUI_LOADFILE:
					DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_OpenFileRequester);
                    break;

                case MP_CMD_GUI_LOADSUBTITLE:
					DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 1, MM_MPlayerGroup_OpenSubtitleRequester);
                    break;

                default:
                    break;
            }
            break;
        }
        case guiSetFileName:
		{
			D(kprintf("guiSetFileName %s\n", arg));
			if (arg) guiIntfStruct.Filename = (char *) arg;
            break;
		}
        case guiSetDefaults:
        {
			D(kprintf("guiSetDefaults\n"));
            audio_id = -1;
            video_id = -1;
            dvdsub_id = -1;
            vobsub_id = -1;
            stream_cache_size = -1;
            autosync = 0;
            vcd_track = 0;
            dvd_title = 0;
            force_fps = 0;

			free(mygui->icy_info);
			mygui->icy_info = NULL;

			if(!mygui->playlist->tracks) return 0;
			guiSetFilename(guiIntfStruct.Filename, mygui->playlist->tracks[mygui->playlist->current]->filename);
			filename = guiIntfStruct.Filename;
            guiIntfStruct.Track = mygui->playlist->current + 1;

            if(gtkAONorm) greplace(&af_cfg.list, "volnorm", "volnorm");
            if(gtkAOExtraStereo)
            {
                char *name = malloc(12 + 20 + 1);
                snprintf(name, 12 + 20, "extrastereo=%f", gtkAOExtraStereoMul);
                name[12 + 20] = 0;
                greplace(&af_cfg.list, "extrastereo", name);
                free(name);
			}

			if(gtkAutoSyncOn) autosync = gtkAutoSync;

            break;
        }
        case guiSetVolume:
        {
            if(audio_out)
            {
                /* Some audio_out drivers do not support balance e.g. dsound */
                /* FIXME this algo is not correct */
                float l, r;
                mixer_getvolume(mixer, &l, &r);
                guiIntfStruct.Volume = (r > l ? r : l); /* max(r,l) */
                if (r != l)
                    guiIntfStruct.Balance = ((r-l) + 100.0f) * 0.5f;
                else
                    guiIntfStruct.Balance = 50.0f;
            }
            break;
        }

		case guiSetWindowPtr:
		{
			D(kprintf("guiSetWindowPtr %p\n", arg));
			if(mygui->embedded)
			{
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_OpenFullScreen, (APTR) arg);
			}

			else
			{
#if !defined(__amigaos4__) /* AROS: Don't put MPlayer window on screen which has video */
				mygui->gui_ready = FALSE;
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_SetWindow, (APTR) arg);
				while(mygui->running && !mygui->gui_ready) // wait until SetWindow is finished or gui is quitting
				{
					Delay(1);
				}
#endif
			}
			break;
		}

		case guiShowPanel:
		{
			D(kprintf("guiShowPanel %p\n", arg));

			gui_show_gui = (ULONG) arg;

			if(!mygui->embedded)
			{
				DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_Show, gui_show_gui);
			}
			else
			{
                DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 2, MM_MPlayerGroup_ShowPanels, gui_show_gui);
			}
			break;
		}

		case guiLoadFile:
		{
			char * path = (char *) arg;
			ULONG previoustrackcount = mygui->playlist->trackcount;

			if(!parse_filename(path, playtree, mconfig, 0))
			{
				mygui->playlist->add_track(mygui->playlist, path, NULL, NULL, 0, STREAMTYPE_FILE);
			}

			mygui->playlist->current = previoustrackcount;
			
			// Rebuild playlist
			update_playlistwindow(TRUE);

			mplSetFileName(NULL, mygui->playlist->tracks[mygui->playlist->current]->filename, STREAMTYPE_FILE);
			mygui->playercontrol(evLoadPlay);

			break;		  
		}

        default:
            mp_msg(MSGT_GPLAYER, MSGL_ERR, "[GUI] GOT UNHANDLED EVENT %i\n", type);
    }
    return 0;
}

/* This function adds/inserts one file into the gui playlist */
int import_file_into_gui(char *pathname, int insert)
{
    char filename[MAX_PATH];
	struct stat st;

	if (strstr(pathname, "dvd://") || strstr(pathname, "dvdnav://"))
	{
		char buffer[MAX_PATH];

		/* Set DVD title here too */
		ULONG dvdnav = strstr(pathname, "dvdnav://") != NULL;

        dvdname[0] = 0;
		buffer[0] = 0;
		dvd_read_name(buffer, NULL, dvd_device);
		capitalize(buffer);
		snprintf(dvdname, sizeof(dvdname), "DVD Movie %s", buffer[0] ? buffer : dvd_device );
		mp_msg(MSGT_GPLAYER, MSGL_INFO, "[GUI] Adding DVD %s\n", dvdname);
		mygui->playlist->add_track(mygui->playlist, pathname, NULL, dvdname, 0, dvdnav ? STREAMTYPE_DVDNAV : STREAMTYPE_DVD);

        return 1;
	}
	else if(strstr(pathname, "://"))
    {
		mp_msg(MSGT_GPLAYER, MSGL_INFO, "[GUI] Adding special %s\n", pathname);
		mygui->playlist->add_track(mygui->playlist, pathname, NULL, pathname, 0, STREAMTYPE_STREAM);

        return 1;
    }

	if(stat(pathname, &st) != -1)
	{
		if (!S_ISDIR(st.st_mode))
	    {
			mp_msg(MSGT_GPLAYER, MSGL_INFO, "[GUI] Adding filename: %s - fullpath: %s\n", FilePart(pathname), pathname);
			mygui->playlist->add_track(mygui->playlist, pathname, NULL, NULL, 0, STREAMTYPE_FILE);

	        return 1;
	    }
	    else
		{
			mp_msg(MSGT_GPLAYER, MSGL_INFO, "[GUI] Cannot add %s\n", filename);
		}
	}

    return 0;
}

/*  This function imports the initial playtree (based on cmd-line files) into the gui playlist
    by either:
    - overwriting gui pl (enqueue=0) */

int import_initial_playtree_into_gui(play_tree_t *my_playtree, m_config_t *config, int enqueue)
{
    play_tree_iter_t *my_pt_iter = NULL;
    int result = 0;

    if(!mygui) guiInit();

	mygui->playlist->clear_playlist(mygui->playlist);

    if((my_pt_iter = pt_iter_create(&my_playtree, config)))
    {
        while ((filename = pt_iter_get_next_file(my_pt_iter)) != NULL)
        {
            if (parse_filename(filename, my_playtree, config, 0))
                result = 1;
            else if (import_file_into_gui(filename, 0)) /* Add it to end of list */
                result = 1;
        }
    }
    mplGotoTheNext = 1;

    if (result)
    {
		mplSetFileName(NULL, mygui->playlist->tracks[0]->filename, mygui->playlist->tracks[0]->type);
		mygui->playlist->current = 0;

		filename = guiIntfStruct.Filename;
    }

	// Rebuild playlist
	update_playlistwindow(TRUE);

    return result;
}

/* This function imports and inserts an playtree, that is created "on the fly", for example by
   parsing some MOV-Reference-File; or by loading an playlist with "File Open"
   The file which contained the playlist is thereby replaced with it's contents. */

int import_playtree_playlist_into_gui(play_tree_t *my_playtree, m_config_t *config)
{
    play_tree_iter_t *my_pt_iter = NULL;
    int result = 0;

    if((my_pt_iter = pt_iter_create(&my_playtree, config)))
    {
        while ((filename = pt_iter_get_next_file(my_pt_iter)) != NULL)
            if (import_file_into_gui(filename, 1)) /* insert it into the list and set plCurrent = new item */
                result = 1;
        pt_iter_destroy(&my_pt_iter);
    }
    filename = NULL;

	// Rebuild playlist
	update_playlistwindow(TRUE);

    return result;
}

void SendMessage(int level, const char *str)
{
	if(mygui)
	{
        static TEXT status[1024];
		static ULONG i = 0;
		char *ptr = str;
		char *dst = status + i;

		while(*ptr && i < sizeof(status))
		{
            if(*ptr == '\n')
			{
				*dst = 0;
				i = 0;
				ptr++;

				// We can send now we have a new line.
				if(mygui->message_pushid)
				{
					DoMethod(app, MUIM_Application_KillPushMethod, mygui->consolegroup, mygui->message_pushid, MM_ConsoleGroup_AddMessage);
				}
				mygui->message_pushid = DoMethod(app, MUIM_Application_PushMethod, mygui->consolegroup, 2, MM_ConsoleGroup_AddMessage, strdup(status));
			}
			else if(*ptr == '\t')
			{
				ptr++;
			}
			else if(*ptr == '\r')
			{
				if(*(ptr + 1) != '\n')
				{
					// We should modify last line content there.
					ptr++;
				}
				else
				{
					ptr++;
				}
			}
			else if(*ptr == 0x1b)
			{
				if(*(ptr + 1))
				{
					ptr++;
					if(*(ptr + 2))
						ptr++;
				}
				ptr++;
			}
			else
			{
				*dst++ = *ptr++;
				i++;
			}
		}
	}
}

void guiMessageBox(int level, char *str)
{
	static TEXT status[256];
    switch(level)
    {
        case MSGL_FATAL:
        case MSGL_ERR:
		case MSGL_INFO:
			SendMessage(level, str);
			break;
		case MSGL_STATUS:
		{
			if(mygui && gui_show_status && str && *str && *str != '\n')
			{
				ULONG secs, micros, currentTime;
				
				CurrentTime(&secs, &micros);

				currentTime = secs*1000 + micros/1000;
				
				// don't refresh too often
				if(currentTime > mygui->last_status_update + 10)
				{
					int i = 0;
					char *ptr = str;
					char *dst = status;
					
					while(*ptr && i < sizeof(status))
					{
						if(*ptr == '\n')
						{	 
							ptr++;
						}
						else if(*ptr == '\r')
						{
							ptr++;
						}
						else if(*ptr == '\t')
						{
							ptr++;
						}
						else if(*ptr == 0x1b) // Skip escape sequences
						{
							if(*(ptr + 1))
							{
								ptr++;
								if(*(ptr + 2))
									ptr++;
							}
							ptr++;
						}
						else
						{
							*dst++ = *ptr;
							i++;
							ptr++;
						}
					}
					*dst = 0;

					if(mygui->status_pushid)
					{
						DoMethod(app, MUIM_Application_KillPushMethod, mygui->maingroup, mygui->status_pushid, MM_MPlayerGroup_Update);
					}

                    mygui->last_status_update = currentTime;
					mygui->status_pushid = DoMethod(app, MUIM_Application_PushMethod, mygui->maingroup, 3, MM_MPlayerGroup_Update, MV_MPlayerGroup_Update_Status, status);
				}
				else
				{
					mygui->last_status_update = currentTime;
				}
			}
			break;
		}
    }
}

/* catch events here in gui mode */
void guiEventHandling(void)
{
#if MPLAYER
	/* REXX PORT */
	gui_handle_events();
	/* END REXX PORT */
#endif
}
